from __future__ import absolute_import, division, print_function

import platform
import sys
from threading import Thread, Lock
import json
import warnings
import time

import fintecture
import pytest

from fintecture.environments import ENVIRONMENT_TEST


if platform.python_implementation() == "PyPy":
    pytest.skip("skip integration tests with PyPy", allow_module_level=True)

if sys.version_info[0] < 3:
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
else:
    from http.server import BaseHTTPRequestHandler, HTTPServer


class TestIntegration(object):
    @pytest.fixture(autouse=True)
    def close_mock_server(self):
        yield
        if self.mock_server:
            self.mock_server.shutdown()
            self.mock_server.server_close()
            self.mock_server_thread.join()

    @pytest.fixture(autouse=True)
    def api_base(self):
        from fintecture import production_api_base, sandbox_api_base, env

        if env not in fintecture.AVAILABLE_ENVS:
            raise ValueError(
                "Defined environment value is invalid. "
                "Please check that specified environment value is one of %r\n" % fintecture.AVAILABLE_ENVS
            )
        if env == fintecture.ENVIRONMENT_SANDBOX:
            return sandbox_api_base
        elif env == fintecture.ENVIRONMENT_PRODUCTION:
            return production_api_base

        return sandbox_api_base

    @pytest.fixture(autouse=True)
    def setup_fintecture(self):
        orig_attrs = {
            "env": fintecture.env,
            "app_id": fintecture.app_id,
            "app_secret": fintecture.app_secret,
            "private_key": fintecture.private_key,
            "default_http_client": fintecture.default_http_client,
            "enable_telemetry": fintecture.enable_telemetry,
            "max_network_retries": fintecture.max_network_retries,
            "proxy": fintecture.proxy,
        }
        fintecture.env = ENVIRONMENT_TEST
        fintecture.app_id = "test_123"
        fintecture.app_secret = "test_456"
        fintecture.private_key = "private_key_789"
        fintecture.default_http_client = None
        fintecture.enable_telemetry = False
        fintecture.max_network_retries = 3
        fintecture.proxy = None
        yield
        fintecture.env = orig_attrs["env"]
        fintecture.app_id = orig_attrs["app_id"]
        fintecture.app_secret = orig_attrs["app_secret"]
        fintecture.private_key = orig_attrs["private_key"]
        fintecture.default_http_client = orig_attrs["default_http_client"]
        fintecture.enable_telemetry = orig_attrs["enable_telemetry"]
        fintecture.max_network_retries = orig_attrs["max_network_retries"]
        fintecture.proxy = orig_attrs["proxy"]

    def setup_mock_server(self, handler):
        # Configure mock server.
        # Passing 0 as the port will cause a random free port to be chosen.
        self.mock_server = HTTPServer(("localhost", 0), handler)
        _, self.mock_server_port = self.mock_server.server_address

        # Start running mock server in a separate thread.
        # Daemon threads automatically shut down when the main process exits.
        self.mock_server_thread = Thread(target=self.mock_server.serve_forever)
        self.mock_server_thread.daemon = True
        self.mock_server_thread.start()

    def test_hits_api_base(self):
        class MockServerRequestHandler(BaseHTTPRequestHandler):
            num_requests = 0

            def do_GET(self):
                self.__class__.num_requests += 1

                self.send_response(200)
                self.send_header(
                    "Content-Type", "application/json; charset=utf-8"
                )
                self.end_headers()
                self.wfile.write(json.dumps({}).encode("utf-8"))
                return

        self.setup_mock_server(MockServerRequestHandler)

        fintecture.api_base = "http://localhost:%s" % self.mock_server_port
        fintecture.Balance.retrieve()
        assert MockServerRequestHandler.num_requests == 1

    def test_hits_proxy_through_default_http_client(self):
        class MockServerRequestHandler(BaseHTTPRequestHandler):
            num_requests = 0

            def do_GET(self):
                self.__class__.num_requests += 1

                self.send_response(200)
                self.send_header(
                    "Content-Type", "application/json; charset=utf-8"
                )
                self.end_headers()
                self.wfile.write(json.dumps({}).encode("utf-8"))
                return

        self.setup_mock_server(MockServerRequestHandler)

        fintecture.proxy = "http://localhost:%s" % self.mock_server_port
        fintecture.Balance.retrieve()
        assert MockServerRequestHandler.num_requests == 1

        fintecture.proxy = "http://bad-url"

        with warnings.catch_warnings(record=True) as w:
            fintecture.Balance.retrieve()
            assert len(w) == 1
            assert "fintecture.proxy was updated after sending a request" in str(
                w[0].message
            )

        assert MockServerRequestHandler.num_requests == 2

    def test_hits_proxy_through_custom_client(self):
        class MockServerRequestHandler(BaseHTTPRequestHandler):
            num_requests = 0

            def do_GET(self):
                self.__class__.num_requests += 1

                self.send_response(200)
                self.send_header(
                    "Content-Type", "application/json; charset=utf-8"
                )
                self.end_headers()
                self.wfile.write(json.dumps({}).encode("utf-8"))
                return

        self.setup_mock_server(MockServerRequestHandler)

        fintecture.default_http_client = (
            fintecture.http_client.new_default_http_client(
                proxy="http://localhost:%s" % self.mock_server_port
            )
        )
        fintecture.Balance.retrieve()
        assert MockServerRequestHandler.num_requests == 1

    def test_passes_client_telemetry_when_enabled(self):
        class MockServerRequestHandler(BaseHTTPRequestHandler):
            num_requests = 0

            def do_GET(self):
                try:
                    self.__class__.num_requests += 1
                    req_num = self.__class__.num_requests
                    if req_num == 1:
                        time.sleep(31 / 1000)  # 31 ms
                        assert not self.headers.get(
                            "X-Fintecture-Client-Telemetry"
                        )
                    elif req_num == 2:
                        assert self.headers.get("X-Fintecture-Client-Telemetry")
                        telemetry = json.loads(
                            self.headers.get("x-Fintecture-client-telemetry")
                        )
                        assert "last_request_metrics" in telemetry
                        req_id = telemetry["last_request_metrics"][
                            "x_request_id"
                        ]
                        duration_ms = telemetry["last_request_metrics"][
                            "request_duration_ms"
                        ]
                        assert req_id == "req_1"
                        # The first request took 31 ms, so the client perceived
                        # latency shouldn't be outside this range.
                        assert 30 < duration_ms < 300
                    else:
                        assert False, (
                            "Should not have reached request %d" % req_num
                        )

                    self.send_response(200)
                    self.send_header(
                        "Content-Type", "application/json; charset=utf-8"
                    )
                    self.send_header("X-Request-ID", "req_%d" % req_num)
                    self.end_headers()
                    self.wfile.write(json.dumps({}).encode("utf-8"))
                except AssertionError as ex:
                    # Throwing assertions on the server side causes a
                    # connection error to be logged instead of an assertion
                    # failure. Instead, we return the assertion failure as
                    # json so it can be logged as a FintectureError.
                    self.send_response(400)
                    self.send_header(
                        "Content-Type", "application/json; charset=utf-8"
                    )
                    self.end_headers()
                    self.wfile.write(
                        json.dumps(
                            {
                                "error": {
                                    "type": "invalid_request_error",
                                    "message": str(ex),
                                }
                            }
                        ).encode("utf-8")
                    )

        self.setup_mock_server(MockServerRequestHandler)
        fintecture.api_base = "http://localhost:%s" % self.mock_server_port
        fintecture.enable_telemetry = True

        fintecture.Balance.retrieve()
        fintecture.Balance.retrieve()
        assert MockServerRequestHandler.num_requests == 2

    def test_uses_thread_local_client_telemetry(self):
        class MockServerRequestHandler(BaseHTTPRequestHandler):
            num_requests = 0
            seen_metrics = set()
            stats_lock = Lock()

            def do_GET(self):
                with self.__class__.stats_lock:
                    self.__class__.num_requests += 1
                    req_num = self.__class__.num_requests

                if self.headers.get("X-Fintecture-Client-Telemetry"):
                    telemetry = json.loads(
                        self.headers.get("X-Fintecture-Client-Telemetry")
                    )
                    req_id = telemetry["last_request_metrics"]["x_request_id"]
                    with self.__class__.stats_lock:
                        self.__class__.seen_metrics.add(req_id)

                self.send_response(200)
                self.send_header(
                    "Content-Type", "application/json; charset=utf-8"
                )
                self.send_header("X-Request-ID", "req_%d" % req_num)
                self.end_headers()
                self.wfile.write(json.dumps({}).encode("utf-8"))

        self.setup_mock_server(MockServerRequestHandler)
        fintecture.api_base = "http://localhost:%s" % self.mock_server_port
        fintecture.enable_telemetry = True
        fintecture.default_http_client = fintecture.http_client.RequestsClient()

        def work():
            fintecture.Balance.retrieve()
            fintecture.Balance.retrieve()

        threads = [Thread(target=work) for _ in range(10)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert MockServerRequestHandler.num_requests == 20
        assert len(MockServerRequestHandler.seen_metrics) == 10
